# @Time : 05/02/2024 16:45
# @Author : a.salgas
# @File : non_recurring_costs.py
# @Software: PyCharm
import numpy as np
import pandas as pd

from aeromaps.models.base import AeroMAPSModel


class NonRecurringCosts(AeroMAPSModel):
    def __init__(self, name="non_recurring_costs", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self.fleet_model = fleet_model

    def compute(
        self,
        aircraft_in_out_value_dict: dict,
    ) -> dict:
        nrc_aircraft_value_dict = {}
        for category, sets in self.fleet_model.fleet.all_aircraft_elements.items():
            # Calculating values of interest for each aircraft
            for aircraft_var in sets:
                # Check if it's a reference aircraft or a normal aircraft...
                if hasattr(aircraft_var, "parameters"):
                    aircraft_var_name = aircraft_var.parameters.full_name
                    nrc_cost = aircraft_var.parameters.nrc_cost
                    eis = int(aircraft_var.parameters.entry_into_service_year)
                else:
                    aircraft_var_name = aircraft_var.full_name
                    nrc_cost = aircraft_var.nrc_cost
                    eis = int(aircraft_var.entry_into_service_year)

                nrc_aircraft_var_name = aircraft_var_name + ":aircraft_non_recurring_costs"

                # TODO use dictionnary if possible once implementeed
                # nrc_aircraft_value = max(0.0, aircraft_in_out_value_dict[aircraft_var_name] * nrc_cost)
                # For now: direct use of fleet model df

                nrc_aircraft_value = self._compute_nrc(float(nrc_cost), 5, eis)

                self.fleet_model.df = pd.concat(
                    [
                        self.fleet_model.df,
                        nrc_aircraft_value.rename(nrc_aircraft_var_name),
                    ],
                    axis=1,
                )

                nrc_aircraft_value_dict[aircraft_var_name] = nrc_aircraft_value

        return nrc_aircraft_value_dict

    def _compute_nrc(
        self,
        nrc_tot_aircraft_type,
        development_time_aircraft_type,
        entry_into_service_year,
    ):
        costs = []
        years = []
        sigma = development_time_aircraft_type / 6.0  # Controls the spread of the distribution

        for i in range(1, development_time_aircraft_type + 1):
            year = entry_into_service_year - development_time_aircraft_type + i
            weight = np.exp(-0.5 * ((i - development_time_aircraft_type / 2) / sigma) ** 2)
            cost = round(nrc_tot_aircraft_type * weight)
            costs.append(cost)
            years.append(year)

        # Adjust the distributed costs to ensure their sum matches the total cost
        total_distributed_cost = sum(costs)
        scaling_factor = nrc_tot_aircraft_type / total_distributed_cost
        costs = [round(cost * scaling_factor) for cost in costs]
        nrc_distributed = pd.Series(costs, index=years)

        # Keep only teh indexes that are within the scope of scenario range to
        # avoid adding out of range nrc (for old aircraft for instance)
        filtered_nrc_distributed = nrc_distributed[
            (nrc_distributed.index >= self.historic_start_year)
            & (nrc_distributed.index <= self.end_year)
        ]

        return filtered_nrc_distributed



# @Time : 13/03/2023 15:14
# @Author : a.salgas
# @File : abatement_potential.py
# @Software: PyCharm

from typing import Tuple

import pandas as pd

from aeromaps.models.base import AeroMAPSModel


class DropinAbatementPotential(AeroMAPSModel):
    def __init__(self, name="dropin_abatement_potential", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        biofuel_atj_efficiency: pd.Series,
        kerosene_emission_factor: pd.Series,
        biofuel_atj_emission_factor: pd.Series,
        biofuel_hefa_fuel_efficiency: pd.Series,
        biofuel_hefa_fog_emission_factor: pd.Series,
        biofuel_hefa_oil_efficiency: pd.Series,
        biofuel_hefa_others_emission_factor: pd.Series,
        biofuel_ft_efficiency: pd.Series,
        biofuel_ft_msw_emission_factor: pd.Series,
        biofuel_ft_others_emission_factor: pd.Series,
        electricity_hydrogen_consumption: pd.Series,
        electrolysis_efficiency: pd.Series,
        electricity_electrofuel_consumption: pd.Series,
        electrofuel_hydrogen_efficiency: pd.Series,
        electrofuel_emission_factor: pd.Series,
        available_biomass_atj: float,
        available_biomass_hefa_fog: float,
        available_biomass_hefa_others: float,
        available_biomass_ft_msw: float,
        available_biomass_ft_others: float,
        aviation_biomass_allocated_share: float,
        aviation_available_electricity: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """Maximal and effective abatement potential through biofuel usage under the allocated biomass hypothesis."""

        ##### Maximal #####

        energy_avail_atj = (
            aviation_biomass_allocated_share
            / 100
            * available_biomass_atj
            * 10**12
            * biofuel_atj_efficiency
        )

        abatement_potential_atj = (
            energy_avail_atj * (kerosene_emission_factor - biofuel_atj_emission_factor) / 1000000
        )

        # Biomass consumption in in EJ, while emission factors are in gCO2e/MJ, abatement potential is in tCO2e

        energy_avail_hefa_fog = (
            aviation_biomass_allocated_share
            / 100
            * available_biomass_hefa_fog
            * 10**12
            * biofuel_hefa_fuel_efficiency
        )
        abatement_potential_hefa_fog = (
            energy_avail_hefa_fog
            * (kerosene_emission_factor - biofuel_hefa_fog_emission_factor)
            / 1000000
        )

        energy_avail_hefa_others = (
            aviation_biomass_allocated_share
            / 100
            * available_biomass_hefa_others
            * 10**12
            * biofuel_hefa_oil_efficiency
            * biofuel_hefa_fuel_efficiency
        )
        abatement_potential_hefa_others = (
            energy_avail_hefa_others
            * (kerosene_emission_factor - biofuel_hefa_others_emission_factor)
            / 1000000
        )

        energy_avail_ft_msw = (
            aviation_biomass_allocated_share
            / 100
            * available_biomass_ft_msw
            * 10**12
            * biofuel_ft_efficiency
        )
        abatement_potential_ft_msw = (
            energy_avail_ft_msw
            * (kerosene_emission_factor - biofuel_ft_msw_emission_factor)
            / 1000000
        )

        energy_avail_ft_others = (
            aviation_biomass_allocated_share
            / 100
            * available_biomass_ft_others
            * 10**12
            * biofuel_ft_efficiency
        )
        abatement_potential_ft_others = (
            energy_avail_ft_others
            * (kerosene_emission_factor - biofuel_ft_others_emission_factor)
            / 1000000
        )

        ##  reactivate for potential hydrogen MACC if ever implemented
        # h2_electrolysis_avoided_emissions_factor = (
        #     kerosene_emission_factor / hydrogen_replacement_ratio
        #     - liquid_hydrogen_electrolysis_emission_factor
        # )
        # energy_avail_hydrogen_electrolysis = (
        #     aviation_available_electricity
        #     * electricity_hydrogen_consumption
        #     / (electricity_electrofuel_consumption + electricity_hydrogen_consumption)
        #     * 10**12
        #     * electrolysis_efficiency
        #     * liquefaction_efficiency
        # )
        # abatement_potential_hydrogen_electrolysis = (
        #     energy_avail_hydrogen_electrolysis * h2_electrolysis_avoided_emissions_factor / 1000000
        # )

        ## The same electricity consumption share (hydrogen/electrofuel) is kept for maxiaml abatement potential computation

        energy_avail_electrofuel = (
            aviation_available_electricity
            * electricity_electrofuel_consumption
            / (electricity_electrofuel_consumption + electricity_hydrogen_consumption)
            * 10**12
            * electrolysis_efficiency
            * electrofuel_hydrogen_efficiency
        )
        abatement_potential_electrofuel = (
            energy_avail_electrofuel
            * (kerosene_emission_factor - electrofuel_emission_factor)
            / 1000000
        )

        self.df.loc[:, "abatement_potential_atj"] = abatement_potential_atj
        self.df.loc[:, "abatement_potential_hefa_fog"] = abatement_potential_hefa_fog
        self.df.loc[:, "abatement_potential_hefa_others"] = abatement_potential_hefa_others
        self.df.loc[:, "abatement_potential_ft_msw"] = abatement_potential_ft_msw
        self.df.loc[:, "abatement_potential_ft_others"] = abatement_potential_ft_others
        # self.df.loc[
        #     :, "abatement_potential_hydrogen_electrolysis"
        # ] = abatement_potential_hydrogen_electrolysis
        self.df.loc[:, "abatement_potential_electrofuel"] = abatement_potential_electrofuel
        self.df.loc[:, "energy_avail_atj"] = energy_avail_atj
        self.df.loc[:, "energy_avail_hefa_fog"] = energy_avail_hefa_fog
        self.df.loc[:, "energy_avail_hefa_others"] = energy_avail_hefa_others
        self.df.loc[:, "energy_avail_ft_msw"] = energy_avail_ft_msw
        self.df.loc[:, "energy_avail_ft_others"] = energy_avail_ft_others
        # self.df.loc[:, "energy_avail_hydrogen_electrolysis"] = energy_avail_hydrogen_electrolysis
        self.df.loc[:, "energy_avail_electrofuel"] = energy_avail_electrofuel

        return (
            abatement_potential_atj,
            abatement_potential_hefa_fog,
            abatement_potential_hefa_others,
            abatement_potential_ft_msw,
            abatement_potential_ft_others,
            abatement_potential_electrofuel,
            energy_avail_atj,
            energy_avail_hefa_fog,
            energy_avail_hefa_others,
            energy_avail_ft_msw,
            energy_avail_ft_others,
            energy_avail_electrofuel,
        )


class EnergyAbatementEffective(AeroMAPSModel):
    def __init__(self, name="energy_abatement_effective", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        biofuel_atj_efficiency: pd.Series,
        biomass_atj_consumption: pd.Series,
        kerosene_emission_factor: pd.Series,
        biofuel_atj_emission_factor: pd.Series,
        biomass_hefa_fog_consumption: pd.Series,
        biofuel_hefa_fuel_efficiency: pd.Series,
        biofuel_hefa_fog_emission_factor: pd.Series,
        biomass_hefa_others_consumption: pd.Series,
        biofuel_hefa_oil_efficiency: pd.Series,
        biofuel_hefa_others_emission_factor: pd.Series,
        biomass_ft_consumption: pd.Series,
        biofuel_ft_efficiency: pd.Series,
        biofuel_ft_msw_emission_factor: pd.Series,
        biofuel_ft_msw_share: pd.Series,
        biofuel_ft_others_share: pd.Series,
        biofuel_ft_others_emission_factor: pd.Series,
        liquid_hydrogen_electrolysis_emission_factor: pd.Series,
        liquid_hydrogen_coal_emission_factor: pd.Series,
        liquid_hydrogen_coal_ccs_emission_factor: pd.Series,
        liquid_hydrogen_gas_emission_factor: pd.Series,
        liquid_hydrogen_gas_ccs_emission_factor: pd.Series,
        energy_consumption_hydrogen: pd.Series,
        hydrogen_electrolysis_share: pd.Series,
        hydrogen_gas_ccs_share: pd.Series,
        hydrogen_coal_ccs_share: pd.Series,
        hydrogen_gas_share: pd.Series,
        hydrogen_coal_share: pd.Series,
        electrolysis_efficiency: pd.Series,
        electricity_electrofuel_consumption: pd.Series,
        electrofuel_hydrogen_efficiency: pd.Series,
        electrofuel_emission_factor: pd.Series,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """Maximal and effective abatement potential through biofuel usage under the allocated biomass hypothesis."""
        ##### Effective #####
        abatement_effective_atj = (
            biomass_atj_consumption
            * 10**12
            * biofuel_atj_efficiency
            * (kerosene_emission_factor - biofuel_atj_emission_factor)
            / 1000000
        )

        # Biomass consumption in in EJ, while emission factors are in gCO2e/MJ, abatement effective is in tCO2e

        abatement_effective_hefa_fog = (
            biomass_hefa_fog_consumption
            * 10**12
            * biofuel_hefa_fuel_efficiency
            * (kerosene_emission_factor - biofuel_hefa_fog_emission_factor)
            / 1000000
        )

        abatement_effective_hefa_others = (
            biomass_hefa_others_consumption
            * 10**12
            * biofuel_hefa_oil_efficiency
            * biofuel_hefa_fuel_efficiency
            * (kerosene_emission_factor - biofuel_hefa_others_emission_factor)
            / 1000000
        )

        abatement_effective_ft_msw = (
            biomass_ft_consumption
            * (biofuel_ft_msw_share / (biofuel_ft_others_share + biofuel_ft_msw_share))
            * 10**12
            * biofuel_ft_efficiency
            * (kerosene_emission_factor - biofuel_ft_msw_emission_factor)
            / 1000000
        )

        abatement_effective_ft_others = (
            biomass_ft_consumption
            * (biofuel_ft_others_share / (biofuel_ft_others_share + biofuel_ft_msw_share))
            * 10**12
            * biofuel_ft_efficiency
            * (kerosene_emission_factor - biofuel_ft_others_emission_factor)
            / 1000000
        )

        #  H2 extra energy handling is supposed to be done in fleet abatement -> energy_replacement_ratio parameter deleted

        h2_electrolysis_avoided_emissions_factor = (
            kerosene_emission_factor - liquid_hydrogen_electrolysis_emission_factor
        )

        abatement_effective_hydrogen_electrolysis = (
            energy_consumption_hydrogen
            * hydrogen_electrolysis_share
            / 100
            * h2_electrolysis_avoided_emissions_factor
            / 1000000
        )

        h2_coal_avoided_emissions_factor = (
            kerosene_emission_factor - liquid_hydrogen_coal_emission_factor
        )

        abatement_effective_hydrogen_coal = (
            energy_consumption_hydrogen
            * hydrogen_coal_share
            / 100
            * h2_coal_avoided_emissions_factor
            / 1000000
        )

        h2_coal_ccs_avoided_emissions_factor = (
            kerosene_emission_factor - liquid_hydrogen_coal_ccs_emission_factor
        )

        abatement_effective_hydrogen_coal_ccs = (
            energy_consumption_hydrogen
            * hydrogen_coal_ccs_share
            / 100
            * h2_coal_ccs_avoided_emissions_factor
            / 1000000
        )

        h2_gas_avoided_emissions_factor = (
            kerosene_emission_factor - liquid_hydrogen_gas_emission_factor
        )

        abatement_effective_hydrogen_gas = (
            energy_consumption_hydrogen
            * hydrogen_gas_share
            / 100
            * h2_gas_avoided_emissions_factor
            / 1000000
        )

        h2_gas_ccs_avoided_emissions_factor = (
            kerosene_emission_factor - liquid_hydrogen_gas_ccs_emission_factor
        )

        abatement_effective_hydrogen_gas_ccs = (
            energy_consumption_hydrogen
            * hydrogen_gas_ccs_share
            / 100
            * h2_gas_ccs_avoided_emissions_factor
            / 1000000
        )

        abatement_effective_electrofuel = (
            electricity_electrofuel_consumption
            * 10**12
            * electrolysis_efficiency
            * electrofuel_hydrogen_efficiency
            * (kerosene_emission_factor - electrofuel_emission_factor)
            / 1000000
        )

        self.df.loc[:, "abatement_effective_atj"] = abatement_effective_atj
        self.df.loc[:, "abatement_effective_hefa_fog"] = abatement_effective_hefa_fog
        self.df.loc[:, "abatement_effective_hefa_others"] = abatement_effective_hefa_others
        self.df.loc[:, "abatement_effective_ft_msw"] = abatement_effective_ft_msw
        self.df.loc[:, "abatement_effective_ft_others"] = abatement_effective_ft_others
        self.df.loc[:, "abatement_effective_hydrogen_electrolysis"] = (
            abatement_effective_hydrogen_electrolysis
        )
        self.df.loc[:, "abatement_effective_hydrogen_coal"] = abatement_effective_hydrogen_coal
        self.df.loc[:, "abatement_effective_hydrogen_coal_ccs"] = (
            abatement_effective_hydrogen_coal_ccs
        )
        self.df.loc[:, "abatement_effective_hydrogen_gas"] = abatement_effective_hydrogen_gas
        self.df.loc[:, "abatement_effective_hydrogen_gas_ccs"] = (
            abatement_effective_hydrogen_gas_ccs
        )
        self.df.loc[:, "abatement_effective_electrofuel"] = abatement_effective_electrofuel

        return (
            abatement_effective_atj,
            abatement_effective_hefa_fog,
            abatement_effective_hefa_others,
            abatement_effective_ft_msw,
            abatement_effective_ft_others,
            abatement_effective_hydrogen_electrolysis,
            abatement_effective_hydrogen_coal,
            abatement_effective_hydrogen_coal_ccs,
            abatement_effective_hydrogen_gas,
            abatement_effective_hydrogen_gas_ccs,
            abatement_effective_electrofuel,
        )



import json
from json import load
import numpy as np
import pandas as pd
from pandas import read_csv
import os.path as pth

from aeromaps.resources import data
from aeromaps.resources import climate_data


def _dict_from_json(file_name="parameters.json") -> dict:
    with open(file_name, "r", encoding="utf-8") as f:
        parameters_dict = load(f)

    for key, value in parameters_dict.items():
        # TODO: generic handling of timetables
        if isinstance(value, list) and key in [
            "rpk_init",
            "ask_init",
            "rtk_init",
            "pax_init",
            "freight_init",
            "energy_consumption_init",
            "total_aircraft_distance_init",
        ]:
            new_index = range(
                parameters_dict["historic_start_year"], parameters_dict["prospection_start_year"]
            )
            parameters_dict[key] = pd.Series(value, index=new_index)

    return parameters_dict


def _dict_to_df(data, orient="index") -> pd.DataFrame:
    # Check if values from data have the same length or else populate with NaN
    max_len = max([len(v) for v in data.values()])
    for key, value in data.items():
        if len(value) < max_len:
            data[key] = np.append(value, np.full(max_len - len(value), np.nan))

    df = pd.DataFrame.from_dict(data, orient=orient)
    return df


def create_partitioning(file, path=""):
    """Generation of a JSON input file (air transport data) and a CSV file (climate data) for running an AeroMAPS process for a partitioned scope."""

    # World input data recovery
    world_data_path = pth.join(data.__path__[0], "parameters.json")
    with open(world_data_path, "r") as parameters_file:
        world_data_dict = json.load(parameters_file)

    # Assumption on freight
    freight_energy_share_2019_partitioned = world_data_dict["freight_energy_share_2019"]

    # AeroSCOPE data recovery
    partitioned_data_df = read_csv(file, delimiter=",")
    partitioned_data = partitioned_data_df.values
    total_ask_2019 = partitioned_data[0, 1]
    short_range_ask_2019 = partitioned_data[0, 2]
    medium_range_ask_2019 = partitioned_data[0, 3]
    long_range_ask_2019 = partitioned_data[0, 4]
    total_seats_2019 = partitioned_data[2, 1]
    total_energy_consumption_per_ask_2019 = partitioned_data[4, 1]
    short_range_energy_consumption_per_ask_2019 = partitioned_data[4, 2]
    medium_range_energy_consumption_per_ask_2019 = partitioned_data[4, 3]
    long_range_energy_consumption_per_ask_2019 = partitioned_data[4, 4]
    total_energy_consumption_2019 = (
        total_energy_consumption_per_ask_2019
        * total_ask_2019
        / (1 - freight_energy_share_2019_partitioned / 2 / 100)
    )  # Dedicated freight (half of total freight) not included in AeroSCOPE
    short_range_energy_consumption_2019 = (
        short_range_energy_consumption_per_ask_2019 * short_range_ask_2019
    ) * (1 - 0.075 / (1 - 0.075))
    medium_range_energy_consumption_2019 = (
        medium_range_energy_consumption_per_ask_2019 * medium_range_ask_2019
    ) * (1 - 0.075 / (1 - 0.075))
    long_range_energy_consumption_2019 = (
        long_range_energy_consumption_per_ask_2019 * long_range_ask_2019
    ) * (1 - 0.075 / (1 - 0.075))

    # Calculation of the partitioned input values

    ## Float inputs
    short_range_energy_share_2019_partitioned = (
        short_range_energy_consumption_2019 / total_energy_consumption_2019 * 100
    )
    medium_range_energy_share_2019_partitioned = (
        medium_range_energy_consumption_2019 / total_energy_consumption_2019 * 100
    )
    long_range_energy_share_2019_partitioned = (
        long_range_energy_consumption_2019 / total_energy_consumption_2019 * 100
    )
    short_range_rpk_share_2019_partitioned = short_range_ask_2019 / total_ask_2019 * 100
    medium_range_rpk_share_2019_partitioned = medium_range_ask_2019 / total_ask_2019 * 100
    long_range_rpk_share_2019_partitioned = long_range_ask_2019 / total_ask_2019 * 100
    commercial_aviation_coefficient_partitioned = 1

    ## Vector inputs
    share_ask_partitioned_vs_world_2019 = total_ask_2019 / world_data_dict["ask_init"][19] * 100
    share_seats_partitioned_vs_world_2019 = total_seats_2019 / (
        world_data_dict["pax_init"][19]
        * world_data_dict["ask_init"][19]
        / world_data_dict["rpk_init"][19]
        * 100
    )
    share_energy_consumption_partitioned_vs_world_2019 = (
        total_energy_consumption_2019 / world_data_dict["energy_consumption_init"][19] * 100
    )
    rpk_init_partitioned = []
    ask_init_partitioned = []
    rtk_init_partitioned = []
    total_aircraft_distance_init_partitioned = []
    freight_init_partitioned = []
    pax_init_partitioned = []
    energy_consumption_init_partitioned = []
    for k in range(0, 20):
        rpk_init_partitioned.append(
            world_data_dict["rpk_init"][k] * share_ask_partitioned_vs_world_2019 / 100
        )
        ask_init_partitioned.append(
            world_data_dict["ask_init"][k] * share_ask_partitioned_vs_world_2019 / 100
        )
        rtk_init_partitioned.append(
            world_data_dict["rtk_init"][k] * share_ask_partitioned_vs_world_2019 / 100
        )
        freight_init_partitioned.append(
            world_data_dict["freight_init"][k] * share_ask_partitioned_vs_world_2019 / 100
        )
        total_aircraft_distance_init_partitioned.append(
            world_data_dict["total_aircraft_distance_init"][k]
            * share_ask_partitioned_vs_world_2019
            / 100
        )
        pax_init_partitioned.append(
            world_data_dict["pax_init"][k] * share_seats_partitioned_vs_world_2019 / 100
        )
        energy_consumption_init_partitioned.append(
            world_data_dict["energy_consumption_init"][k]
            * share_energy_consumption_partitioned_vs_world_2019
            / 100
        )

    # TODO move historic and prospection start year out of custom input file

    historic_start_year_partitioned = world_data_dict["historic_start_year"]
    prospection_start_year_partitioned = world_data_dict["prospection_start_year"]

    # Generation of the JSON file
    partitioned_inputs_dict = {
        "rpk_init": rpk_init_partitioned,
        "ask_init": ask_init_partitioned,
        "rtk_init": rtk_init_partitioned,
        "pax_init": pax_init_partitioned,
        "freight_init": freight_init_partitioned,
        "energy_consumption_init": energy_consumption_init_partitioned,
        "total_aircraft_distance_init": total_aircraft_distance_init_partitioned,
        "short_range_energy_share_2019": short_range_energy_share_2019_partitioned,
        "medium_range_energy_share_2019": medium_range_energy_share_2019_partitioned,
        "long_range_energy_share_2019": long_range_energy_share_2019_partitioned,
        "freight_energy_share_2019": freight_energy_share_2019_partitioned,
        "short_range_rpk_share_2019": short_range_rpk_share_2019_partitioned,
        "medium_range_rpk_share_2019": medium_range_rpk_share_2019_partitioned,
        "long_range_rpk_share_2019": long_range_rpk_share_2019_partitioned,
        "commercial_aviation_coefficient": commercial_aviation_coefficient_partitioned,
        "historic_start_year": historic_start_year_partitioned,
        "prospection_start_year": prospection_start_year_partitioned,
    }
    partitioned_inputs_path = pth.join(path, "partitioned_inputs.json")
    with open(partitioned_inputs_path, "w") as outfile:
        json.dump(partitioned_inputs_dict, outfile)

    # Generation of a CSV file for using climate models
    climate_world_data_path = pth.join(
        climate_data.__path__[0], "temperature_historical_dataset.csv"
    )
    climate_world_data_df = pd.read_csv(climate_world_data_path, delimiter=";")
    climate_world_data = climate_world_data_df.values
    climate_world_data_years = climate_world_data[:, 0]
    climate_world_data_co2_emissions = climate_world_data[:, 1]
    climate_world_data_nox_emissions = climate_world_data[:, 2]
    climate_world_data_h2o_emissions = climate_world_data[:, 3]
    climate_world_data_soot_emissions = climate_world_data[:, 4]
    climate_world_data_sulfur_emissions = climate_world_data[:, 5]
    climate_world_data_distance = climate_world_data[:, 6]
    climate_partitioned_data_years = climate_world_data_years
    climate_partitioned_data_co2_emissions = (
        climate_world_data_co2_emissions * share_energy_consumption_partitioned_vs_world_2019 / 100
    )
    climate_partitioned_data_nox_emissions = (
        climate_world_data_nox_emissions * share_energy_consumption_partitioned_vs_world_2019 / 100
    )
    climate_partitioned_data_h2o_emissions = (
        climate_world_data_h2o_emissions * share_energy_consumption_partitioned_vs_world_2019 / 100
    )
    climate_partitioned_data_soot_emissions = (
        climate_world_data_soot_emissions * share_energy_consumption_partitioned_vs_world_2019 / 100
    )
    climate_partitioned_data_sulfur_emissions = (
        climate_world_data_sulfur_emissions
        * share_energy_consumption_partitioned_vs_world_2019
        / 100
    )
    climate_partitioned_data_distance = (
        climate_world_data_distance * share_ask_partitioned_vs_world_2019 / 100
    )
    climate_partitioned_data_years_number = len(climate_partitioned_data_years)
    partitioned_historical_climate_dataset = np.zeros((climate_partitioned_data_years_number, 7))
    for k in range(0, climate_partitioned_data_years_number):
        partitioned_historical_climate_dataset[k, 0] = climate_partitioned_data_years[k]
        partitioned_historical_climate_dataset[k, 1] = climate_partitioned_data_co2_emissions[k]
        partitioned_historical_climate_dataset[k, 2] = climate_partitioned_data_nox_emissions[k]
        partitioned_historical_climate_dataset[k, 3] = climate_partitioned_data_h2o_emissions[k]
        partitioned_historical_climate_dataset[k, 4] = climate_partitioned_data_soot_emissions[k]
        partitioned_historical_climate_dataset[k, 5] = climate_partitioned_data_sulfur_emissions[k]
        partitioned_historical_climate_dataset[k, 6] = climate_partitioned_data_distance[k]
    climate_partitioned_data_path = pth.join(path, "partitioned_temperature_historical_dataset.csv")
    np.savetxt(climate_partitioned_data_path, partitioned_historical_climate_dataset, delimiter=";")

    return



# Contains the principal high-level functions of AeroMAPS
from aeromaps.core.models import default_models_top_down
from aeromaps.core.process import AeroMAPSProcess


def create_process(
    configuration_file=None,
    models=default_models_top_down,
    use_fleet_model=False,
    add_examples_aircraft_and_subcategory=True,
) -> AeroMAPSProcess:
    """
    Create an AeroMAPS process.
    """

    return AeroMAPSProcess(
        configuration_file=configuration_file,
        models=models,
        use_fleet_model=use_fleet_model,
        add_examples_aircraft_and_subcategory=add_examples_aircraft_and_subcategory,
    )



# @Time : 24/02/2023 17:44
# @Author : a.salgas
# @File : power_to_liquid.py
# @Software: PyCharm

from typing import Tuple

import numpy as np
import pandas as pd

from aeromaps.models.base import AeroMAPSModel, AeromapsInterpolationFunction


class ElectrofuelCost(AeroMAPSModel):
    def __init__(self, name="electrofuel_cost", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_electrofuel: pd.Series,
        electrofuel_eis_capex: pd.Series,
        electrofuel_eis_fixed_opex: pd.Series,
        electrofuel_eis_var_opex: pd.Series,
        electrolysis_efficiency: pd.Series,
        electrofuel_hydrogen_efficiency: pd.Series,
        kerosene_market_price: pd.Series,
        kerosene_emission_factor: pd.Series,
        electrofuel_emission_factor: pd.Series,
        electricity_market_price: pd.Series,
        co2_market_price: pd.Series,
        electrofuel_eis_specific_co2: pd.Series,
        electricity_load_factor: pd.Series,
        carbon_tax: pd.Series,
        exogenous_carbon_price_trajectory: pd.Series,
        plant_lifespan: float,
        private_discount_rate: float,
        social_discount_rate: float,
        lhv_electrofuel: float,
        density_electrofuel: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        ######## HYDROGEN PRODUCTION ########

        #### ELECTROLYSIS ####

        (
            electrofuel_plant_building_scenario,
            electrofuel_plant_building_cost,
            electrofuel_total_cost,
            electrofuel_mean_capex_share,
            electrofuel_mean_opex_share,
            electrofuel_mean_elec_share,
            electrofuel_mean_co2_share,
            electrofuel_mean_mfsp_litre,
            electrofuel_cost_premium,
            electrofuel_carbon_tax,
            electrofuel_mfsp_carbon_tax_supplement,
            carbon_abatement_cost_electrofuel,
            specific_carbon_abatement_cost_electrofuel,
            generic_specific_carbon_abatement_cost_electrofuel,
        ) = self._electrofuel_computation(
            electrofuel_eis_capex,
            electrofuel_eis_fixed_opex,
            electrofuel_eis_var_opex,
            electrolysis_efficiency,
            electrofuel_hydrogen_efficiency,
            electricity_market_price,
            energy_consumption_electrofuel,
            electricity_load_factor,
            co2_market_price,
            electrofuel_eis_specific_co2,
            kerosene_emission_factor,
            electrofuel_emission_factor,
            kerosene_market_price,
            carbon_tax,
            exogenous_carbon_price_trajectory,
            plant_lifespan,
            private_discount_rate,
            social_discount_rate,
            lhv_electrofuel,
            density_electrofuel,
        )

        self.df.loc[:, "electrofuel_plant_building_scenario"] = electrofuel_plant_building_scenario
        self.df.loc[:, "electrofuel_plant_building_cost"] = electrofuel_plant_building_cost
        self.df.loc[:, "electrofuel_total_cost"] = electrofuel_total_cost
        self.df.loc[:, "electrofuel_mean_capex_share"] = electrofuel_mean_capex_share
        self.df.loc[:, "electrofuel_mean_opex_share"] = electrofuel_mean_opex_share
        self.df.loc[:, "electrofuel_mean_elec_share"] = electrofuel_mean_elec_share
        self.df.loc[:, "electrofuel_mean_co2_share"] = electrofuel_mean_co2_share
        self.df.loc[:, "electrofuel_cost_premium"] = electrofuel_cost_premium
        self.df.loc[:, "electrofuel_mean_mfsp_litre"] = electrofuel_mean_mfsp_litre
        self.df.loc[:, "carbon_abatement_cost_electrofuel"] = carbon_abatement_cost_electrofuel
        self.df.loc[:, "specific_carbon_abatement_cost_electrofuel"] = (
            specific_carbon_abatement_cost_electrofuel
        )
        self.df.loc[:, "generic_specific_carbon_abatement_cost_electrofuel"] = (
            generic_specific_carbon_abatement_cost_electrofuel
        )
        self.df.loc[:, "electrofuel_carbon_tax"] = electrofuel_carbon_tax
        self.df.loc[:, "electrofuel_mfsp_carbon_tax_supplement"] = (
            electrofuel_mfsp_carbon_tax_supplement
        )

        return (
            electrofuel_plant_building_scenario,
            electrofuel_plant_building_cost,
            electrofuel_total_cost,
            electrofuel_mean_capex_share,
            electrofuel_mean_opex_share,
            electrofuel_mean_elec_share,
            electrofuel_mean_co2_share,
            electrofuel_cost_premium,
            electrofuel_mean_mfsp_litre,
            carbon_abatement_cost_electrofuel,
            electrofuel_carbon_tax,
            electrofuel_mfsp_carbon_tax_supplement,
            specific_carbon_abatement_cost_electrofuel,
            generic_specific_carbon_abatement_cost_electrofuel,
        )

    def _electrofuel_computation(
        self,
        electrofuel_eis_capex: pd.Series,
        electrofuel_eis_fixed_opex: pd.Series,
        electrofuel_eis_var_opex: pd.Series,
        # electrofuel_eis_specific_energy: pd.Series,
        electrolysis_efficiency: pd.Series,
        electrofuel_hydrogen_efficiency: pd.Series,
        electricity_market_price: pd.Series,
        energy_consumption_electrofuel: pd.Series,
        electricity_load_factor: pd.Series,
        co2_market_price: pd.Series,
        electrofuel_eis_specific_co2: pd.Series,
        kerosene_emission_factor: pd.Series,
        electrofuel_emission_factor: pd.Series,
        kerosene_market_price: pd.Series,
        carbon_tax: pd.Series,
        exogenous_carbon_price_trajectory: pd.Series,
        plant_lifespan: float,
        private_discount_rate: float,
        social_discount_rate: float,
        lhv_electrofuel: float,
        density_electrofuel: float,
    ):
        """
        Computes the yearly costs to respect a given hydrogen electrolysis production scenario.
        Capex in €/ton/day
        Fixed opex in  €/ton/day
        Variable opex in € per kg
        Specific electricity in kWh/kg --> Deprecated to harmonize with efficiency compytation, but might be reactivated later?
        Electricity market price in €/kWh
        Energy_consumption_hydrogen in MJ.
        Values returned are in M€ or t/day
        """

        # Constants:
        construction_time = 3
        plant_load_fact = 0.95

        # Avoided emission factor in gCO2/MJ
        avoided_emission_factor = kerosene_emission_factor - electrofuel_emission_factor

        # Demand scenario for the pathway in MJ
        demand_scenario = energy_consumption_electrofuel

        # Catching the indexes from the demand scenario
        indexes = demand_scenario.index

        # Additional plant capacity building needed (in ton/day output size) for a given year
        plant_building_scenario = pd.Series(np.zeros(len(indexes)), indexes)
        # Relative CAPEX cost to build the new facilities in M€2020
        plant_building_cost = pd.Series(np.zeros(len(indexes)), indexes)

        # Annual production in MJ
        electrofuel_production = pd.Series(np.zeros(len(indexes)), indexes)

        # carbon abatement cost in €/ton
        # carbon_abatement_cost = pd.Series(np.zeros(len(indexes)), indexes)

        specific_carbon_abatement_cost = pd.Series(np.nan, indexes)
        generic_specific_carbon_abatement_cost = pd.Series(np.nan, indexes)

        # Annual cost and cost components of hydrogen production in M€
        electrofuel_total_cost = pd.Series(np.zeros(len(indexes)), indexes)
        electrofuel_capex_cost = pd.Series(np.zeros(len(indexes)), indexes)
        electrofuel_opex_cost = pd.Series(np.zeros(len(indexes)), indexes)
        electrofuel_elec_cost = pd.Series(np.zeros(len(indexes)), indexes)
        electrofuel_co2_cost = pd.Series(np.zeros(len(indexes)), indexes)

        # Total extra cost linked to carbon tax in M€2020
        # electrofuel_carbon_tax_cost = pd.Series(np.zeros(len(indexes)), indexes)
        # Total annual cost premium in M€2020
        # electrofuel_cost_premium = pd.Series(np.zeros(len(indexes)), indexes)

        # Extra cost on mfsp linked to carbon tax in €/L
        # mfsp_supplement_carbon_tax = pd.Series(np.zeros(len(indexes)), indexes)

        # For each year o²f the demand scenario the demand is matched by the production
        for year in list(demand_scenario.index):
            # Production missing in year n+1 must be supplied by plant build in year n
            if (year + 1) <= self.end_year and electrofuel_production[year + 1] < demand_scenario[
                year + 1
            ]:
                # Getting the production not matched by plants already commissioned by creating an electrolyzer with year data technical data

                electrofuel_cost = self._compute_electrofuel_year_lcop(
                    int(construction_time),
                    int(plant_lifespan),
                    year - construction_time,
                    electricity_market_price,
                    co2_market_price,
                    electricity_load_factor,
                    electrofuel_eis_capex,
                    electrofuel_eis_fixed_opex,
                    electrofuel_eis_var_opex,
                    electrolysis_efficiency,
                    electrofuel_hydrogen_efficiency,
                    electrofuel_eis_specific_co2,
                    private_discount_rate,
                    plant_load_fact,
                    lhv_electrofuel,
                    density_electrofuel,
                )

                # Getting the production not matched by plants already commissioned
                missing_production = demand_scenario[year + 1] - electrofuel_production[year + 1]

                # Converting the missing production to a capacity (in t/day)
                missing_production_kg = missing_production / lhv_electrofuel
                missing_production_litres = missing_production_kg / density_electrofuel
                electrofuel_capacity_to_build = (
                    missing_production_kg
                    / 365.25
                    / min(electricity_load_factor[year], plant_load_fact)
                )  # capacity to build in kg/day production, taking into account load_factor

                electrolyser_capex_year = (
                    electrofuel_capacity_to_build * electrofuel_eis_capex[year] / 1000000
                )  # electrolyzer capex is in €/kg/day => M€
                plant_building_scenario[year] = electrofuel_capacity_to_build  # in ton/day capacity

                for construction_year in range(year - construction_time, year):
                    if self.historic_start_year < construction_year < self.end_year:
                        plant_building_cost[construction_year] += (
                            electrolyser_capex_year / construction_time
                        )

                # When production ends: either at the end of plant life or the end of the scenario;
                end_bound = int(min(list(demand_scenario.index)[-1], year + plant_lifespan))

                discounted_cumul_cost = 0
                cumul_em = 0
                generic_discounted_cumul_em = 0

                # Adding new plant production to future years
                for i in range(year + 1, end_bound + 1):
                    electrofuel_total_cost[i] = (
                        electrofuel_total_cost[i]
                        + (missing_production_litres * electrofuel_cost[i]["TOTAL"]) / 1000000
                    )  # €/L and production in litres => /1000000 for M€
                    electrofuel_capex_cost[i] = (
                        electrofuel_capex_cost[i]
                        + (missing_production_litres * electrofuel_cost[i]["CAPEX"]) / 1000000
                    )  # M€
                    electrofuel_opex_cost[i] = (
                        electrofuel_opex_cost[i]
                        + (missing_production_litres * electrofuel_cost[i]["FIX_OPEX"]) / 1000000
                        + (missing_production_litres * electrofuel_cost[i]["VAR_OPEX"]) / 1000000
                    )  # M€
                    electrofuel_elec_cost[i] = (
                        electrofuel_elec_cost[i]
                        + (missing_production_litres * electrofuel_cost[i]["ELECTRICITY"]) / 1000000
                    )  # M€
                    electrofuel_co2_cost[i] = (
                        electrofuel_co2_cost[i]
                        + (missing_production_litres * electrofuel_cost[i]["CO2"]) / 1000000
                    )  # M€
                    electrofuel_production[i] = electrofuel_production[i] + missing_production

                for i in range(year, year + int(plant_lifespan)):
                    if i < (self.end_year + 1):
                        discounted_cumul_cost += (
                            electrofuel_cost[i]["TOTAL"] - kerosene_market_price[i]
                        ) / (1 + social_discount_rate) ** (i - year)
                        cumul_em += (
                            avoided_emission_factor[i]
                            * (lhv_electrofuel * density_electrofuel)
                            / 1000000
                        )

                        # discounting emissions for non-hotelling scc
                        generic_discounted_cumul_em += (
                            avoided_emission_factor[i]
                            * (lhv_electrofuel * density_electrofuel)
                            / 1000000
                            * exogenous_carbon_price_trajectory[i]
                            / exogenous_carbon_price_trajectory[year]
                            / (1 + social_discount_rate) ** (i - year)
                        )
                    else:
                        discounted_cumul_cost += (
                            electrofuel_cost[self.end_year]["TOTAL"]
                            - kerosene_market_price[self.end_year]
                        ) / (1 + social_discount_rate) ** (i - year)
                        cumul_em += (
                            avoided_emission_factor[self.end_year]
                            * (lhv_electrofuel * density_electrofuel)
                            / 1000000
                        )

                        # discounting emissions for non-hotelling scc, keep last year scc growth rate as future scc growth rate
                        future_scc_growth = (
                            exogenous_carbon_price_trajectory[self.end_year]
                            / exogenous_carbon_price_trajectory[self.end_year - 1]
                        )

                        generic_discounted_cumul_em += (
                            avoided_emission_factor[self.end_year]
                            * (lhv_electrofuel * density_electrofuel)
                            / 1000000
                            * (
                                exogenous_carbon_price_trajectory[self.end_year]
                                / exogenous_carbon_price_trajectory[year]
                                * (future_scc_growth) ** (i - self.end_year)
                            )
                            / (1 + social_discount_rate) ** (i - year)
                        )

                # Using unitary values for cost and emission possible as long as the plant operates at constant capacity during its life
                # (Volume gets out of cac sums)
                specific_carbon_abatement_cost[year] = discounted_cumul_cost / cumul_em
                generic_specific_carbon_abatement_cost[year] = (
                    discounted_cumul_cost / generic_discounted_cumul_em
                )

            elif (year == self.end_year) or (
                electrofuel_production[year + 1] >= demand_scenario[year + 1] > 0
            ):
                specific_carbon_abatement_cost[year] = specific_carbon_abatement_cost[year - 1]
                generic_specific_carbon_abatement_cost[year] = (
                    generic_specific_carbon_abatement_cost[year - 1]
                )

        # MOD -> Scaling down production for diminishing production scenarios.
        # Very weak model, assuming that production not anymore needed by aviation is used elsewhere in the industry.
        # Stranded asset literature could be valuable to model this better.
        # Proportional production scaling

        scaling_factor = demand_scenario / electrofuel_production

        if not all(scaling_factor.isna()):
            electrofuel_total_cost = electrofuel_total_cost * scaling_factor
            electrofuel_capex_cost = electrofuel_capex_cost * scaling_factor
            electrofuel_opex_cost = electrofuel_opex_cost * scaling_factor
            electrofuel_elec_cost = electrofuel_elec_cost * scaling_factor
            electrofuel_co2_cost = electrofuel_co2_cost * scaling_factor

        electrofuel_mean_mfsp_litre = (
            electrofuel_total_cost
            / (demand_scenario / (lhv_electrofuel * density_electrofuel))
            * 1000000
        )

        electrofuel_mean_capex_share = electrofuel_capex_cost / electrofuel_total_cost * 100
        electrofuel_mean_opex_share = electrofuel_opex_cost / electrofuel_total_cost * 100
        electrofuel_mean_elec_share = electrofuel_elec_cost / electrofuel_total_cost * 100
        electrofuel_mean_co2_share = electrofuel_co2_cost / electrofuel_total_cost * 100

        electrofuel_cost_premium = (
            (electrofuel_mean_mfsp_litre - kerosene_market_price)
            * (demand_scenario / (lhv_electrofuel * density_electrofuel))
            / 1000000
        )

        # Compute the carbon tax (M€)
        electrofuel_carbon_tax = (
            carbon_tax * electrofuel_emission_factor * demand_scenario / 1000000 / 1000000
        )

        electrofuel_mfsp_carbon_tax_supplement = (
            carbon_tax
            * electrofuel_emission_factor
            / 1000000
            * (lhv_electrofuel * density_electrofuel)
        )
        # Abatement cost in €/tCO2e (= overcost for a ton of biofuel/avoided emissions)
        carbon_abatement_cost_electrofuel = (
            (electrofuel_mean_mfsp_litre - kerosene_market_price)
            / (avoided_emission_factor * (lhv_electrofuel * density_electrofuel))
            * 1000000
        )

        return (
            plant_building_scenario,
            plant_building_cost,
            electrofuel_total_cost,
            electrofuel_mean_capex_share,
            electrofuel_mean_opex_share,
            electrofuel_mean_elec_share,
            electrofuel_mean_co2_share,
            electrofuel_mean_mfsp_litre,
            electrofuel_cost_premium,
            electrofuel_carbon_tax,
            electrofuel_mfsp_carbon_tax_supplement,
            carbon_abatement_cost_electrofuel,
            specific_carbon_abatement_cost,
            generic_specific_carbon_abatement_cost,
        )

    def _compute_electrofuel_year_lcop(
        self,
        construction_time,
        plant_lifespan,
        base_year,
        electricity_market_price,
        co2_market_price,
        electricity_load_factor,
        electrofuel_capex,
        electrofuel_fixed_opex,
        electrofuel_var_opex,
        # electrofuel_specific_electricity,
        electrolysis_efficiency,
        electrofuel_hydrogen_efficiency,
        electrofuel_specific_co2,
        private_discount_rate,
        plant_load_fact,
        lhv_electrofuel,
        density_electrofuel,
    ):
        """
        This function computes the MFSP for electrofuel production for an plant commissioned at the base year.
        Costs are discounted to find a constant MFSP, excepted electricity, whose price is directly evolving
        from a year to another. Hydrogen production and fuel synthesis are considered a one-step process here.
        Capex in €/ (kg/day capacity)
        Fixed opex in €/ (kg/day capacity), on an annual basis
        Variable opex in € per L
        Specific electricity in kWh/L
        Specific CO2 in kg/L
        Electricity market price in €/kWh
        Electrofuel price returned in €/L
        """
        hydrogen_prices = {}

        technology_year = max(base_year, self.prospection_start_year)

        load_fact = min(plant_load_fact, electricity_load_factor[technology_year])

        real_year_days = 365.25 * load_fact
        real_var_opex = electrofuel_var_opex[technology_year] * real_year_days

        dropin_specific_energy = lhv_electrofuel * density_electrofuel / 3.6  # kWh/L

        # Electrolysis efficiency correction to remove DAC input from efficiency (Fig.2 of https://www.nature.com/articles/s41558-021-01032-7)

        electrofuel_efficiency = (
            electrolysis_efficiency
            * electrofuel_hydrogen_efficiency
            * (0.81 + 0.03 + 0.06)
            / (0.81 + 0.03)
        )

        electrofuel_specific_electricity = dropin_specific_energy / (electrofuel_efficiency)

        cap_cost_npv = 0
        fix_op_cost_npv = 0
        var_op_cost_npv = 0
        hydrogen_npv = 0

        # Construction
        for i in range(0, construction_time):
            # The construction is supposed to span over x years, with a uniform cost repartition
            cap_cost_npv += (
                electrofuel_capex[technology_year] * density_electrofuel / construction_time
            ) / (1 + private_discount_rate) ** (i)

        # commissioning
        for i in range(construction_time, plant_lifespan + construction_time):
            fix_op_cost_npv += (
                electrofuel_fixed_opex[technology_year]
                * density_electrofuel
                / (1 + private_discount_rate) ** (i)
            )
            var_op_cost_npv += real_var_opex / (1 + private_discount_rate) ** (i)
            hydrogen_npv += real_year_days / (1 + private_discount_rate) ** (i)

        cap_cost_lc = cap_cost_npv / hydrogen_npv
        fix_op_cost_lc = fix_op_cost_npv / hydrogen_npv
        var_op_cost_lc = var_op_cost_npv / hydrogen_npv

        end_bound = min(
            max(electricity_market_price.index), plant_lifespan + construction_time + base_year
        )

        for year in range(base_year + construction_time, int(end_bound) + 1):
            elec_price = electricity_market_price[year]
            elec_cost = elec_price * electrofuel_specific_electricity[technology_year]
            CO2_cost = electrofuel_specific_co2[technology_year] * co2_market_price[year]
            hydrogen_prices[year] = {
                "TOTAL": cap_cost_lc + var_op_cost_lc + fix_op_cost_lc + elec_cost + CO2_cost,
                "CAPEX": cap_cost_lc,
                "FIX_OPEX": fix_op_cost_lc,
                "VAR_OPEX": var_op_cost_lc,
                "ELECTRICITY": elec_cost,
                "CO2": CO2_cost,
            }

        return hydrogen_prices


class ElectrofuelCapex(AeroMAPSModel):
    def __init__(self, name="electrofuel_capex", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_capex_reference_years: list,
        electrofuel_capex_reference_years_values: list,
    ) -> pd.Series:
        """Electrofuel capital expenditures at eis using interpolation functions"""

        electrofuel_eis_capex = AeromapsInterpolationFunction(
            self,
            electrofuel_capex_reference_years,
            electrofuel_capex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "electrofuel_eis_capex"] = electrofuel_eis_capex

        return electrofuel_eis_capex


class ElectrofuelFixedOpex(AeroMAPSModel):
    def __init__(self, name="electrofuel_fixed_opex", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_fixed_opex_reference_years: list,
        electrofuel_fixed_opex_reference_years_values: list,
    ) -> pd.Series:
        """Electrofuel fixed operational expenditures at entry into service using interpolation functions"""

        electrofuel_eis_fixed_opex = AeromapsInterpolationFunction(
            self,
            electrofuel_fixed_opex_reference_years,
            electrofuel_fixed_opex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "electrofuel_eis_fixed_opex"] = electrofuel_eis_fixed_opex

        return electrofuel_eis_fixed_opex


class ElectrofuelVarOpex(AeroMAPSModel):
    def __init__(self, name="electrofuel_var_opex", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_var_opex_reference_years: list,
        electrofuel_var_opex_reference_years_values: list,
    ) -> pd.Series:
        """Electrofuel variable operational expenditures at entry into service using interpolation functions"""

        electrofuel_eis_var_opex = AeromapsInterpolationFunction(
            self,
            electrofuel_var_opex_reference_years,
            electrofuel_var_opex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "electrofuel_eis_var_opex"] = electrofuel_eis_var_opex

        return electrofuel_eis_var_opex


########## Deprecated for the time being, might be reactivated. ####################
#
# class ElectrofuelSpecificElectricity(AeroMAPSModel):
#
#     # changement d'usage par rapport à CAST==> on utilise pas l'efficacité moyenne pour la cons d'élec, mais
#     # l'efficacité de chaque année de mise en service de l'eclectolyseur. Permet de faire des choses plus détaillées
#     #
#     def __init__(self, name="electrofuel_specific_electricity", *args, **kwargs):
#         super().__init__(name, *args, **kwargs)
#
#     def compute(
#             self,
#             electrofuel_specific_electricity_2020: float,
#             electrofuel_specific_electricity_2030: float,
#             electrofuel_specific_electricity_2040: float,
#             electrofuel_specific_electricity_2050: float,
#     ) -> pd.Series:
#         """electrofuel efficiency at eis using interpolation functions"""
#         # FT MSW
#         reference_values_specific_electricity = [
#             electrofuel_specific_electricity_2020,
#             electrofuel_specific_electricity_2030,
#             electrofuel_specific_electricity_2040,
#             electrofuel_specific_electricity_2050
#         ]
#
#         reference_years = [2020, 2030, 2040, self.end_year]
#
#         specific_electricity_function = interp1d(
#             reference_years, reference_values_specific_electricity, kind="linear"
#         )
#         for k in range(self.prospection_start_year, self.end_year + 1):
#             self.df.loc[
#                 k, "electrofuel_eis_specific_electricity"
#             ] = specific_electricity_function(k)
#
#         electrofuel_eis_specific_electricity = self.df.loc[:, "electrofuel_eis_specific_electricity"]
#
#         return (
#             electrofuel_eis_specific_electricity
#         )
####################################################################################


class ElectrofuelSpecificCo2(AeroMAPSModel):
    # changement d'usage par rapport à CAST==> on utilise pas l'efficacité moyenne pour la cons d'élec, mais
    # l'efficacité de chaque année de mise en service de l'eclectolyseur. Permet de faire des choses plus détaillées.
    def __init__(self, name="electrofuel_specific_co2", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_specific_co2_reference_years: list,
        electrofuel_specific_co2_reference_years_values: list,
    ) -> pd.Series:
        """Electrofuel efficiency at eis using interpolation functions"""

        electrofuel_eis_specific_co2 = AeromapsInterpolationFunction(
            self,
            electrofuel_specific_co2_reference_years,
            electrofuel_specific_co2_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "electrofuel_eis_specific_co2"] = electrofuel_eis_specific_co2

        return electrofuel_eis_specific_co2



# @Time : 06/02/2024 15:12
# @Author : a.salgas
# @File : operations_abatement_cost.py
# @Software: PyCharm
from typing import Tuple

import numpy as np
import pandas as pd
from aeromaps.models.base import AeroMAPSModel


class OperationsAbatementCost(AeroMAPSModel):
    def __init__(self, name="operations_abatement_cost", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        operational_efficiency_cost_non_energy_per_ask: pd.Series,
        operations_gain: pd.Series,
        kerosene_market_price: pd.Series,
        kerosene_emission_factor: pd.Series,
        energy_per_ask_mean_without_operations: pd.Series,
        energy_per_ask_mean: pd.Series,
        energy_per_rtk_mean_without_operations: pd.Series,
        energy_per_rtk_mean: pd.Series,
        rpk: pd.Series,
        rtk: pd.Series,
        load_factor: pd.Series,
        load_factor_cost_non_energy_per_ask: pd.Series,
        exogenous_carbon_price_trajectory: pd.Series,
        social_discount_rate: float,
        operations_duration: float,
        operations_start_year: int,
        lhv_kerosene: float,
        density_kerosene: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        fuel_lhv = lhv_kerosene * density_kerosene

        ############### PASSENGER OPERATIONS #############

        emissions_reduction_operations = (
            energy_per_ask_mean_without_operations
            * operations_gain
            / 100
            * kerosene_emission_factor
            / 1000000
        )

        extra_cost_operations_non_fuel = operational_efficiency_cost_non_energy_per_ask

        extra_cost_operations_fuel = (
            -energy_per_ask_mean_without_operations
            * operations_gain
            / 100
            * kerosene_market_price
            / fuel_lhv
        )

        operations_abatement_cost = (
            extra_cost_operations_non_fuel + extra_cost_operations_fuel
        ) / emissions_reduction_operations

        self.df.loc[:, "operations_abatement_cost"] = operations_abatement_cost

        operations_abatement_effective = (
            emissions_reduction_operations
            * rpk
            / load_factor.loc[self.prospection_start_year - 1]
            * 100
        )
        self.df.loc[:, "operations_abatement_effective"] = operations_abatement_effective

        # Definition of a specific abatement cost, comparable to a hotelling growth carbon value.
        # Discount the costs/benefits over the horizon necessary to deploy the incremental gains of a year
        for k in range(int(operations_start_year), self.end_year + 1):
            (scac, scac_prime) = self._get_discounted_vals(
                k,
                social_discount_rate,
                operations_duration,
                extra_cost_operations_non_fuel,
                extra_cost_operations_fuel,
                kerosene_market_price,
                kerosene_emission_factor,
                emissions_reduction_operations,
                exogenous_carbon_price_trajectory,
            )

            self.df.loc[k, "operations_specific_abatement_cost"] = scac
            self.df.loc[k, "operations_generic_specific_abatement_cost"] = scac_prime

        operations_specific_abatement_cost = self.df["operations_specific_abatement_cost"]
        operations_generic_specific_abatement_cost = self.df[
            "operations_generic_specific_abatement_cost"
        ]

        ################ FREIGHT OPERATIONS ################

        emissions_reduction_operations_freight = (
            energy_per_rtk_mean_without_operations
            * operations_gain
            / 100
            * kerosene_emission_factor
            / 1000000
        )

        extra_cost_operations_non_fuel_freight = pd.Series(
            np.zeros(len(self.df.index)), index=self.df.index
        )

        extra_cost_operations_fuel_freight = (
            -energy_per_rtk_mean_without_operations
            * operations_gain
            / 100
            * kerosene_market_price
            / fuel_lhv
        )

        operations_abatement_cost_freight = (
            extra_cost_operations_non_fuel_freight + extra_cost_operations_fuel_freight
        ) / emissions_reduction_operations_freight

        self.df.loc[:, "operations_abatement_cost_freight"] = operations_abatement_cost_freight

        operations_abatement_effective_freight = emissions_reduction_operations_freight * rtk
        self.df.loc[:, "operations_abatement_effective_freight"] = (
            operations_abatement_effective_freight
        )

        # Definition of a specific abatement cost, comparable to a hotelling growth carbon value.
        # Discount the costs/benefits over the horizon necessary to deploy the incremental gains of a year
        for k in range(int(operations_start_year), self.end_year + 1):
            (scac, scac_prime) = self._get_discounted_vals(
                k,
                social_discount_rate,
                operations_duration,
                extra_cost_operations_non_fuel_freight,
                extra_cost_operations_fuel_freight,
                kerosene_market_price,
                kerosene_emission_factor,
                emissions_reduction_operations_freight,
                exogenous_carbon_price_trajectory,
            )

            self.df.loc[k, "operations_specific_abatement_cost_freight"] = scac
            self.df.loc[k, "operations_generic_specific_abatement_cost_freight"] = scac_prime

        operations_specific_abatement_cost_freight = self.df[
            "operations_specific_abatement_cost_freight"
        ]
        operations_generic_specific_abatement_cost_freight = self.df[
            "operations_generic_specific_abatement_cost_freight"
        ]

        ############### PASSENGER LOAD FACTOR ########################

        energy_per_rpk_base = (
            energy_per_ask_mean / load_factor.loc[self.prospection_start_year - 1] * 100
        )
        energy_per_rpk_real = energy_per_ask_mean / load_factor * 100

        emissions_reduction_load_factor = (
            (energy_per_rpk_base - energy_per_rpk_real) * kerosene_emission_factor / 1000000
        )
        extra_cost_load_factor_fuel = (
            -(energy_per_rpk_base - energy_per_rpk_real) * kerosene_market_price / fuel_lhv
        )
        extra_cost_load_factor_non_fuel = load_factor_cost_non_energy_per_ask

        load_factor_abatement_cost = (
            extra_cost_load_factor_fuel + extra_cost_load_factor_non_fuel
        ) / emissions_reduction_load_factor

        self.df.loc[:, "load_factor_abatement_cost"] = load_factor_abatement_cost

        load_factor_abatement_effective = emissions_reduction_load_factor * rpk
        self.df.loc[:, "load_factor_abatement_effective"] = load_factor_abatement_effective

        # Definition of a specific abatement cost, comparable to a hotelling growth carbon value.
        # Discount the costs/benefits over the scenario temporal span.
        # Caution: the longer the scenario, the lower the specific abatement cost
        for k in range(self.prospection_start_year, self.end_year + 1):
            scac, scac_prime = self._get_discounted_vals(
                k,
                social_discount_rate,
                self.end_year - self.prospection_start_year,
                extra_cost_load_factor_non_fuel,
                extra_cost_load_factor_fuel,
                kerosene_market_price,
                kerosene_emission_factor,
                emissions_reduction_load_factor,
                exogenous_carbon_price_trajectory,
            )

            self.df.loc[k, "load_factor_specific_abatement_cost"] = scac
            self.df.loc[k, "load_factor_generic_specific_abatement_cost"] = scac_prime

        load_factor_specific_abatement_cost = self.df["load_factor_specific_abatement_cost"]
        load_factor_generic_specific_abatement_cost = self.df[
            "load_factor_generic_specific_abatement_cost"
        ]

        return (
            operations_abatement_cost,
            operations_abatement_effective,
            operations_specific_abatement_cost,
            operations_generic_specific_abatement_cost,
            operations_abatement_cost_freight,
            operations_abatement_effective_freight,
            operations_specific_abatement_cost_freight,
            operations_generic_specific_abatement_cost_freight,
            load_factor_abatement_cost,
            load_factor_abatement_effective,
            load_factor_specific_abatement_cost,
            load_factor_generic_specific_abatement_cost,
        )

    def _get_discounted_vals(
        self,
        year,
        discount_rate,
        measure_duration,
        extra_cost_non_fuel,
        extra_cost_fuel,
        kerosene_market_price,
        kerosene_emission_factor,
        emissions_reduction,
        exogenous_carbon_price_trajectory,
    ):
        discounted_cumul_cost = 0
        cumul_em = 0
        generic_discounted_cumul_em = 0
        for i in range(year, year + int(measure_duration)):
            if i < (self.end_year + 1):
                discounted_cumul_cost += (
                    extra_cost_non_fuel[year]
                    + extra_cost_fuel[year] * kerosene_market_price[i] / kerosene_market_price[year]
                ) / (1 + discount_rate) ** (i - year)
                cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[i]
                    / kerosene_emission_factor[year]
                )

                generic_discounted_cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[i]
                    / kerosene_emission_factor[year]
                    * exogenous_carbon_price_trajectory[i]
                    / exogenous_carbon_price_trajectory[year]
                    / (1 + discount_rate) ** (i - year)
                )
            else:
                discounted_cumul_cost += (
                    extra_cost_non_fuel[year]
                    + extra_cost_fuel[year]
                    * kerosene_market_price[self.end_year]
                    / kerosene_market_price[year]
                ) / (1 + discount_rate) ** (i - year)
                cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[self.end_year]
                    / kerosene_emission_factor[year]
                )

                # discounting emissions for non-hotelling scc, keep last year scc growth rate as future scc growth rate
                future_scc_growth = (
                    exogenous_carbon_price_trajectory[self.end_year]
                    / exogenous_carbon_price_trajectory[self.end_year - 1]
                )

                generic_discounted_cumul_em += (
                    emissions_reduction[year]
                    * (kerosene_emission_factor[self.end_year] / kerosene_emission_factor[year])
                    * (
                        exogenous_carbon_price_trajectory[self.end_year]
                        / exogenous_carbon_price_trajectory[year]
                        * (future_scc_growth) ** (i - self.end_year)
                    )
                    / (1 + discount_rate) ** (i - year)
                )

        return (
            discounted_cumul_cost / cumul_em,
            discounted_cumul_cost / generic_discounted_cumul_em,
        )



# @Time : 05/02/2024 16:23
# @Author : a.salgas
# @File : fleet_abatement_cost.py
# @Software: PyCharm
import pandas as pd

from aeromaps.models.base import AeroMAPSModel
from typing import Tuple


class FleetCarbonAbatementCosts(AeroMAPSModel):
    def __init__(self, name="fleet_abatement_cost", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self.fleet_model = fleet_model

    def compute(
        self,
        ask_aircraft_value_dict: dict,
        rpk_aircraft_value_dict: dict,
        load_factor: pd.Series,
        doc_non_energy_per_ask_short_range_dropin_fuel_init: float,
        doc_non_energy_per_ask_medium_range_dropin_fuel_init: float,
        doc_non_energy_per_ask_long_range_dropin_fuel_init: float,
        energy_per_ask_without_operations_short_range_dropin_fuel: pd.Series,
        energy_per_ask_without_operations_medium_range_dropin_fuel: pd.Series,
        energy_per_ask_without_operations_long_range_dropin_fuel: pd.Series,
        kerosene_market_price: pd.Series,
        kerosene_emission_factor: pd.Series,
        exogenous_carbon_price_trajectory: pd.Series,
        covid_energy_intensity_per_ask_increase_2020: float,
        lhv_kerosene: float,
        density_kerosene: float,
        social_discount_rate: float,
    ) -> Tuple[dict, dict, dict, dict]:
        dummy_carbon_abatement_cost_aircraft_value_dict = {}
        dummy_specific_carbon_abatement_cost_aircraft_value_dict = {}
        dummy_generic_specific_carbon_abatement_cost_aircraft_value_dict = {}
        dummy_carbon_abatement_volume_aircraft_value_dict = {}
        for category, sets in self.fleet_model.fleet.all_aircraft_elements.items():
            category_recent_reference = self.fleet_model.fleet.all_aircraft_elements[category][1]

            if category == "Short Range":
                category_reference_doc_ne = doc_non_energy_per_ask_short_range_dropin_fuel_init
                # Assumes 100% drop in at reference year. Sum with non drop in necessary otherwise.
                category_reference_energy = (
                    energy_per_ask_without_operations_short_range_dropin_fuel[
                        self.prospection_start_year - 1
                    ]
                )
            elif category == "Medium Range":
                category_reference_doc_ne = doc_non_energy_per_ask_medium_range_dropin_fuel_init
                category_reference_energy = (
                    energy_per_ask_without_operations_medium_range_dropin_fuel[
                        self.prospection_start_year - 1
                    ]
                )
            else:
                category_reference_doc_ne = doc_non_energy_per_ask_long_range_dropin_fuel_init
                category_reference_energy = (
                    energy_per_ask_without_operations_long_range_dropin_fuel[
                        self.prospection_start_year - 1
                    ]
                )

            # Calculating values of interest for each aircraft
            for aircraft_var in sets:
                # Check if it's a reference aircraft or a normal aircraft...
                if hasattr(aircraft_var, "parameters"):
                    aircraft_var_name = aircraft_var.parameters.full_name
                    aircraft_energy_delta_val = (
                        1 + float(aircraft_var.parameters.consumption_evolution) / 100
                    ) * category_recent_reference.energy_per_ask - category_reference_energy

                    aircraft_doc_ne_delta = (
                        1 + float(aircraft_var.parameters.doc_non_energy_evolution) / 100
                    ) * category_recent_reference.doc_non_energy_base - category_reference_doc_ne

                else:
                    aircraft_var_name = aircraft_var.full_name
                    aircraft_energy_delta_val = (
                        aircraft_var.energy_per_ask - category_reference_energy
                    )
                    aircraft_doc_ne_delta = (
                        aircraft_var.doc_non_energy_base - category_reference_doc_ne
                    )

                # include covid energy per ask increase, applied indifferently to the whole fleet.
                # Initialisation of a uniform pd.series and modification of 2020 value.
                aircraft_energy_delta = pd.Series(
                    aircraft_energy_delta_val, index=self.fleet_model.df.index
                )

                # initialisation based on the same model transition than aircraft efficiency model, transitioning

                aircraft_energy_delta[2019] = 0  # start from reference

                aircraft_energy_delta[2020] = (
                    aircraft_energy_delta[2019]
                    * (1 + covid_energy_intensity_per_ask_increase_2020 / 100)
                    + category_reference_energy * covid_energy_intensity_per_ask_increase_2020 / 100
                )

                # Assumption: 100% kerosene for cost calculation. Effect of SAFs/Hydrogen is accounted for downwards in
                # the calculation process. For instance a Hydrogen aircraft, that would consume more energy in this step
                # would result in negative abatement; even though the total lifecycle could actually reduce emissions.

                extra_cost_fuel = (
                    aircraft_energy_delta
                    * kerosene_market_price
                    / (lhv_kerosene * density_kerosene)
                )

                extra_emissions = (-aircraft_energy_delta * kerosene_emission_factor) / 1000000

                aircraft_carbon_abatement_cost = (
                    extra_cost_fuel + aircraft_doc_ne_delta
                ) / extra_emissions  # €/ton

                specific_cac_aircraft_var_name = (
                    aircraft_var_name + ":aircraft_specific_carbon_abatement_cost"
                )

                generic_specific_cac_aircraft_var_name = (
                    aircraft_var_name + ":aircraft_generic_specific_carbon_abatement_cost"
                )

                scac_vals = []
                scac_vals_prime = []
                for k in range(self.prospection_start_year, self.end_year + 1):
                    scac, scac_prime = self._get_discounted_vals(
                        k,
                        social_discount_rate,
                        self.fleet_model.fleet.categories[category].parameters.life,
                        aircraft_doc_ne_delta,
                        extra_cost_fuel,
                        kerosene_market_price,
                        kerosene_emission_factor,
                        extra_emissions,
                        exogenous_carbon_price_trajectory,
                    )
                    scac_vals.append(scac)
                    scac_vals_prime.append(scac_prime)

                scac_column = pd.DataFrame(
                    {specific_cac_aircraft_var_name: scac_vals},
                    index=range(self.prospection_start_year, self.end_year + 1),
                )

                scac_prime_column = pd.DataFrame(
                    {generic_specific_cac_aircraft_var_name: scac_vals_prime},
                    index=range(self.prospection_start_year, self.end_year + 1),
                )

                self.fleet_model.df = pd.concat([self.fleet_model.df, scac_column], axis=1)
                self.fleet_model.df = pd.concat([self.fleet_model.df, scac_prime_column], axis=1)

                # TODO use input dictionary if possible once implemented instead of looking in fleet model df + dummy output
                aircraft_pseudo_ask = (
                    self.fleet_model.df.loc[:, (aircraft_var_name + ":aircraft_rpk")]
                    / load_factor[self.prospection_start_year - 1]
                    * 100
                )

                aircraft_carbon_abatement_volume = -(
                    aircraft_pseudo_ask * aircraft_energy_delta * kerosene_emission_factor / 1000000
                )  # in tons

                cac_aircraft_var_name = aircraft_var_name + ":aircraft_carbon_abatement_cost"

                abatement_volume_aircraft_var_name = (
                    aircraft_var_name + ":aircraft_carbon_abatement_volume"
                )

                self.fleet_model.df = pd.concat(
                    [
                        self.fleet_model.df,
                        aircraft_carbon_abatement_cost.rename(cac_aircraft_var_name),
                    ],
                    axis=1,
                )

                self.fleet_model.df = pd.concat(
                    [
                        self.fleet_model.df,
                        aircraft_carbon_abatement_volume.rename(abatement_volume_aircraft_var_name),
                    ],
                    axis=1,
                )

                dummy_carbon_abatement_cost_aircraft_value_dict[aircraft_var_name] = (
                    aircraft_carbon_abatement_cost
                )
                dummy_carbon_abatement_volume_aircraft_value_dict[aircraft_var_name] = (
                    aircraft_carbon_abatement_volume
                )

                dummy_specific_carbon_abatement_cost_aircraft_value_dict[aircraft_var_name] = (
                    scac_column
                )

                dummy_generic_specific_carbon_abatement_cost_aircraft_value_dict[
                    aircraft_var_name
                ] = scac_prime_column

        return (
            dummy_carbon_abatement_cost_aircraft_value_dict,
            dummy_carbon_abatement_volume_aircraft_value_dict,
            dummy_specific_carbon_abatement_cost_aircraft_value_dict,
            dummy_generic_specific_carbon_abatement_cost_aircraft_value_dict,
        )

    def _get_discounted_vals(
        self,
        year,
        discount_rate,
        aircraft_life,
        extra_cost_non_fuel,
        extra_cost_fuel,
        kerosene_market_price,
        kerosene_emission_factor,
        emissions_reduction,
        exogenous_carbon_price_trajectory,
    ):
        discounted_cumul_cost = 0
        cumul_em = 0
        generic_discounted_cumul_em = 0
        for i in range(year, year + int(aircraft_life)):
            if i < (self.end_year + 1):
                discounted_cumul_cost += (
                    extra_cost_non_fuel
                    + extra_cost_fuel[year] * kerosene_market_price[i] / kerosene_market_price[year]
                ) / (1 + discount_rate) ** (i - year)
                cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[i]
                    / kerosene_emission_factor[year]
                )

                # discounting emissions for non-hotelling scc
                generic_discounted_cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[i]
                    / kerosene_emission_factor[year]
                    * exogenous_carbon_price_trajectory[i]
                    / exogenous_carbon_price_trajectory[year]
                    / (1 + discount_rate) ** (i - year)
                )
            else:
                discounted_cumul_cost += (
                    extra_cost_non_fuel
                    + extra_cost_fuel[year]
                    * kerosene_market_price[self.end_year]
                    / kerosene_market_price[year]
                ) / (1 + discount_rate) ** (i - year)
                cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[self.end_year]
                    / kerosene_emission_factor[year]
                )

                # discounting emissions for non-hotelling scc, keep last year scc growth rate as future scc growth rate
                future_scc_growth = (
                    exogenous_carbon_price_trajectory[self.end_year]
                    / exogenous_carbon_price_trajectory[self.end_year - 1]
                )

                generic_discounted_cumul_em += (
                    emissions_reduction[year]
                    * (kerosene_emission_factor[self.end_year] / kerosene_emission_factor[year])
                    * (
                        exogenous_carbon_price_trajectory[self.end_year]
                        / exogenous_carbon_price_trajectory[year]
                        * (future_scc_growth) ** (i - self.end_year)
                    )
                    / (1 + discount_rate) ** (i - year)
                )

        return (
            discounted_cumul_cost / cumul_em,
            discounted_cumul_cost / generic_discounted_cumul_em,
        )


class CargoEfficiencyCarbonAbatementCosts(AeroMAPSModel):
    def __init__(
        self, name="cargo_efficiency_carbon_abatement_cost", fleet_model=None, *args, **kwargs
    ):
        super().__init__(name, *args, **kwargs)
        self.fleet_model = fleet_model

    def compute(
        self,
        energy_per_rtk_without_operations_freight_dropin_fuel: pd.Series,
        energy_per_rtk_without_operations_freight_hydrogen: pd.Series,
        energy_per_rtk_without_operations_freight_electric: pd.Series,
        rtk_dropin_fuel: pd.Series,
        rtk_hydrogen: pd.Series,
        rtk_electric: pd.Series,
        kerosene_market_price: pd.Series,
        kerosene_emission_factor: pd.Series,
        exogenous_carbon_price_trajectory: pd.Series,
        lhv_kerosene: float,
        density_kerosene: float,
        social_discount_rate: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        ### WARNING => Cargo DOC are not modelled so far in AeroMAPS (April 2024).
        # However, energy abatement are computed based on energy volumes used by the whole fleet.
        # For consistency, we compute a cargo abatement cost based on the hypothesis that
        # non-energy DOC remain constant, thus not requiring their knowledge.

        freight_reference_energy = energy_per_rtk_without_operations_freight_dropin_fuel.loc[
            self.prospection_start_year - 1
        ]

        category_reference_doc_ne = 0

        aircraft_energy_delta_dropin = (
            energy_per_rtk_without_operations_freight_dropin_fuel - freight_reference_energy
        )
        aircraft_energy_delta_hydrogen = (
            energy_per_rtk_without_operations_freight_hydrogen - freight_reference_energy
        )
        aircraft_energy_delta_electric = (
            energy_per_rtk_without_operations_freight_electric - freight_reference_energy
        )

        aircraft_doc_ne_delta_dropin = 0 - category_reference_doc_ne
        aircraft_doc_ne_delta_hydrogen = 0 - category_reference_doc_ne
        aircraft_doc_ne_delta_electric = 0 - category_reference_doc_ne

        # Assumption: 100% kerosene for cost calculation. Effect of SAFs/Hydrogen is accounted for downwards in
        # the calculation process. For instance a Hydrogen aircraft, that would consume more energy in this step
        # would result in negative abatement; even though the total lifecycle could actually reduce emissions.

        extra_cost_fuel_dropin = (
            aircraft_energy_delta_dropin * kerosene_market_price / (lhv_kerosene * density_kerosene)
        )
        extra_cost_fuel_hydrogen = (
            aircraft_energy_delta_hydrogen
            * kerosene_market_price
            / (lhv_kerosene * density_kerosene)
        )
        extra_cost_fuel_electric = (
            aircraft_energy_delta_electric
            * kerosene_market_price
            / (lhv_kerosene * density_kerosene)
        )

        extra_emissions_dropin = (
            -aircraft_energy_delta_dropin * kerosene_emission_factor
        ) / 1000000
        extra_emissions_hydrogen = (
            -aircraft_energy_delta_hydrogen * kerosene_emission_factor
        ) / 1000000
        extra_emissions_electric = (
            -aircraft_energy_delta_electric * kerosene_emission_factor
        ) / 1000000

        aircraft_carbon_abatement_cost_freight_dropin = (
            extra_cost_fuel_dropin + aircraft_doc_ne_delta_dropin
        ) / extra_emissions_dropin  # €/ton
        aircraft_carbon_abatement_cost_freight_hydrogen = (
            extra_cost_fuel_hydrogen + aircraft_doc_ne_delta_hydrogen
        ) / extra_emissions_hydrogen  # €/ton
        aircraft_carbon_abatement_cost_freight_electric = (
            extra_cost_fuel_electric + aircraft_doc_ne_delta_electric
        ) / extra_emissions_electric  # €/ton

        self.df.loc[:, "aircraft_carbon_abatement_cost_freight_dropin"] = (
            aircraft_carbon_abatement_cost_freight_dropin
        )
        self.df.loc[:, "aircraft_carbon_abatement_cost_freight_hydrogen"] = (
            aircraft_carbon_abatement_cost_freight_hydrogen
        )
        self.df.loc[:, "aircraft_carbon_abatement_cost_freight_electric"] = (
            aircraft_carbon_abatement_cost_freight_electric
        )

        # aircraft_life = (self.fleet_model.fleet.categories['Short Range'].parameters.life +
        #                  self.fleet_model.fleet.categories['Medium Range'].parameters.life +
        #                  self.fleet_model.fleet.categories['Long Range'].parameters.life) / 3

        aircraft_life = 25

        scac_vals_dropin = []
        scac_vals_prime_dropin = []
        for k in range(self.prospection_start_year, self.end_year + 1):
            scac, scac_prime = self._get_discounted_vals(
                k,
                social_discount_rate,
                aircraft_life,
                aircraft_doc_ne_delta_dropin,
                extra_cost_fuel_dropin,
                kerosene_market_price,
                kerosene_emission_factor,
                extra_emissions_dropin,
                exogenous_carbon_price_trajectory,
            )
            scac_vals_dropin.append(scac)
            scac_vals_prime_dropin.append(scac_prime)
        aircraft_specific_carbon_abatement_cost_freight_dropin = pd.Series(
            scac_vals_dropin,
            index=range(self.prospection_start_year, self.end_year + 1),
            name="aircraft_specific_carbon_abatement_cost_freight_dropin",
        )
        aircraft_generic_specific_carbon_abatement_cost_freight_dropin = pd.Series(
            scac_vals_prime_dropin,
            index=range(self.prospection_start_year, self.end_year + 1),
            name="aircraft_generic_specific_carbon_abatement_cost_freight_dropin",
        )
        self.df = pd.concat(
            [self.df, aircraft_specific_carbon_abatement_cost_freight_dropin], axis=1
        )
        self.df = pd.concat(
            [self.df, aircraft_generic_specific_carbon_abatement_cost_freight_dropin], axis=1
        )

        scac_vals_hydrogen = []
        scac_vals_prime_hydrogen = []
        for k in range(self.prospection_start_year, self.end_year + 1):
            scac, scac_prime = self._get_discounted_vals(
                k,
                social_discount_rate,
                aircraft_life,
                aircraft_doc_ne_delta_hydrogen,
                extra_cost_fuel_hydrogen,
                kerosene_market_price,
                kerosene_emission_factor,
                extra_emissions_hydrogen,
                exogenous_carbon_price_trajectory,
            )
            scac_vals_hydrogen.append(scac)
            scac_vals_prime_hydrogen.append(scac_prime)
        aircraft_specific_carbon_abatement_cost_freight_hydrogen = pd.Series(
            scac_vals_hydrogen,
            index=range(self.prospection_start_year, self.end_year + 1),
            name="aircraft_specific_carbon_abatement_cost_freight_hydrogen",
        )
        aircraft_generic_specific_carbon_abatement_cost_freight_hydrogen = pd.Series(
            scac_vals_prime_hydrogen,
            index=range(self.prospection_start_year, self.end_year + 1),
            name="aircraft_generic_specific_carbon_abatement_cost_freight_hydrogen",
        )
        self.df = pd.concat(
            [self.df, aircraft_specific_carbon_abatement_cost_freight_hydrogen], axis=1
        )
        self.df = pd.concat(
            [self.df, aircraft_generic_specific_carbon_abatement_cost_freight_hydrogen], axis=1
        )

        scac_vals_electric = []
        scac_vals_prime_electric = []
        for k in range(self.prospection_start_year, self.end_year + 1):
            scac, scac_prime = self._get_discounted_vals(
                k,
                social_discount_rate,
                aircraft_life,
                aircraft_doc_ne_delta_electric,
                extra_cost_fuel_electric,
                kerosene_market_price,
                kerosene_emission_factor,
                extra_emissions_electric,
                exogenous_carbon_price_trajectory,
            )
            scac_vals_electric.append(scac)
            scac_vals_prime_electric.append(scac_prime)
        aircraft_specific_carbon_abatement_cost_freight_electric = pd.Series(
            scac_vals_electric,
            index=range(self.prospection_start_year, self.end_year + 1),
            name="aircraft_specific_carbon_abatement_cost_freight_electric",
        )
        aircraft_generic_specific_carbon_abatement_cost_freight_electric = pd.Series(
            scac_vals_prime_electric,
            index=range(self.prospection_start_year, self.end_year + 1),
            name="aircraft_generic_specific_carbon_abatement_cost_freight_electric",
        )
        self.df = pd.concat(
            [self.df, aircraft_specific_carbon_abatement_cost_freight_electric], axis=1
        )
        self.df = pd.concat(
            [self.df, aircraft_generic_specific_carbon_abatement_cost_freight_electric], axis=1
        )

        aircraft_carbon_abatement_volume_freight_dropin = -(
            rtk_dropin_fuel * aircraft_energy_delta_dropin * kerosene_emission_factor / 1000000
        )
        aircraft_carbon_abatement_volume_freight_hydrogen = -(
            rtk_hydrogen * aircraft_energy_delta_hydrogen * kerosene_emission_factor / 1000000
        )
        aircraft_carbon_abatement_volume_freight_electric = -(
            rtk_electric * aircraft_energy_delta_electric * kerosene_emission_factor / 1000000
        )

        self.df.loc[:, "aircraft_carbon_abatement_volume_freight_dropin"] = (
            aircraft_carbon_abatement_volume_freight_dropin
        )
        self.df.loc[:, "aircraft_carbon_abatement_volume_freight_hydrogen"] = (
            aircraft_carbon_abatement_volume_freight_hydrogen
        )
        self.df.loc[:, "aircraft_carbon_abatement_volume_freight_electric"] = (
            aircraft_carbon_abatement_volume_freight_electric
        )

        return (
            aircraft_carbon_abatement_cost_freight_dropin,
            aircraft_carbon_abatement_cost_freight_hydrogen,
            aircraft_carbon_abatement_cost_freight_electric,
            aircraft_generic_specific_carbon_abatement_cost_freight_dropin,
            aircraft_generic_specific_carbon_abatement_cost_freight_hydrogen,
            aircraft_generic_specific_carbon_abatement_cost_freight_electric,
            aircraft_specific_carbon_abatement_cost_freight_dropin,
            aircraft_specific_carbon_abatement_cost_freight_hydrogen,
            aircraft_specific_carbon_abatement_cost_freight_electric,
            aircraft_carbon_abatement_volume_freight_dropin,
            aircraft_carbon_abatement_volume_freight_hydrogen,
            aircraft_carbon_abatement_volume_freight_electric,
        )

    def _get_discounted_vals(
        self,
        year,
        discount_rate,
        aircraft_life,
        extra_cost_non_fuel,
        extra_cost_fuel,
        kerosene_market_price,
        kerosene_emission_factor,
        emissions_reduction,
        exogenous_carbon_price_trajectory,
    ):
        discounted_cumul_cost = 0
        cumul_em = 0
        generic_discounted_cumul_em = 0
        for i in range(year, year + int(aircraft_life)):
            if i < (self.end_year + 1):
                discounted_cumul_cost += (
                    extra_cost_non_fuel
                    + extra_cost_fuel[year] * kerosene_market_price[i] / kerosene_market_price[year]
                ) / (1 + discount_rate) ** (i - year)
                cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[i]
                    / kerosene_emission_factor[year]
                )

                # discounting emissions for non-hotelling scc
                generic_discounted_cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[i]
                    / kerosene_emission_factor[year]
                    * exogenous_carbon_price_trajectory[i]
                    / exogenous_carbon_price_trajectory[year]
                    / (1 + discount_rate) ** (i - year)
                )
            else:
                discounted_cumul_cost += (
                    extra_cost_non_fuel
                    + extra_cost_fuel[year]
                    * kerosene_market_price[self.end_year]
                    / kerosene_market_price[year]
                ) / (1 + discount_rate) ** (i - year)
                cumul_em += (
                    emissions_reduction[year]
                    * kerosene_emission_factor[self.end_year]
                    / kerosene_emission_factor[year]
                )

                # discounting emissions for non-hotelling scc, keep last year scc growth rate as future scc growth rate
                future_scc_growth = (
                    exogenous_carbon_price_trajectory[self.end_year]
                    / exogenous_carbon_price_trajectory[self.end_year - 1]
                )

                generic_discounted_cumul_em += (
                    emissions_reduction[year]
                    * (kerosene_emission_factor[self.end_year] / kerosene_emission_factor[year])
                    * (
                        exogenous_carbon_price_trajectory[self.end_year]
                        / exogenous_carbon_price_trajectory[year]
                        * (future_scc_growth) ** (i - self.end_year)
                    )
                    / (1 + discount_rate) ** (i - year)
                )

        return (
            discounted_cumul_cost / cumul_em,
            discounted_cumul_cost / generic_discounted_cumul_em,
        )



# @Time : 22/02/2023 16:01
# @Author : a.salgas
# @File : biofuel.py
# @Software: PyCharm
from typing import Tuple

import pandas as pd
from aeromaps.models.base import AeroMAPSModel, AeromapsInterpolationFunction


class BiofuelCostSimple(AeroMAPSModel):
    def __init__(self, name="biofuel_cost_simple", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_biofuel: pd.Series,
        biofuel_hefa_fog_emission_factor: pd.Series,
        biofuel_hefa_others_emission_factor: pd.Series,
        biofuel_atj_emission_factor: pd.Series,
        biofuel_ft_msw_emission_factor: pd.Series,
        biofuel_ft_others_emission_factor: pd.Series,
        biofuel_hefa_fog_share: pd.Series,
        biofuel_hefa_others_share: pd.Series,
        biofuel_atj_share: pd.Series,
        biofuel_ft_msw_share: pd.Series,
        biofuel_ft_others_share: pd.Series,
        biofuel_hefa_fog_mfsp: pd.Series,
        biofuel_hefa_others_mfsp: pd.Series,
        biofuel_ft_others_mfsp: pd.Series,
        biofuel_ft_msw_mfsp: pd.Series,
        biofuel_atj_mfsp: pd.Series,
        carbon_tax: pd.Series,
        lhv_biofuel: float,
        density_biofuel: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        ### HEFA FOG

        biofuel_cost_hefa_fog = (
            biofuel_hefa_fog_mfsp
            / density_biofuel
            / lhv_biofuel
            * energy_consumption_biofuel
            * biofuel_hefa_fog_share
            / 100
            / 1e6
        )
        self.df.loc[:, "biofuel_cost_hefa_fog"] = biofuel_cost_hefa_fog

        biofuel_carbon_tax_hefa_fog = (
            energy_consumption_biofuel
            * biofuel_hefa_fog_share
            / 100
            * biofuel_hefa_fog_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        self.df.loc[:, "biofuel_carbon_tax_hefa_fog"] = biofuel_carbon_tax_hefa_fog

        biofuel_mfsp_carbon_tax_supplement_hefa_fog = (
            carbon_tax * biofuel_hefa_fog_emission_factor / 1000000 * lhv_biofuel * density_biofuel
        )
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_hefa_fog"] = (
            biofuel_mfsp_carbon_tax_supplement_hefa_fog
        )

        ### HEFA OTHERS
        biofuel_cost_hefa_others = (
            biofuel_hefa_others_mfsp
            / density_biofuel
            / lhv_biofuel
            * energy_consumption_biofuel
            * biofuel_hefa_others_share
            / 100
            / 1e6
        )
        self.df.loc[:, "biofuel_cost_hefa_others"] = biofuel_cost_hefa_others

        biofuel_carbon_tax_hefa_others = (
            energy_consumption_biofuel
            * biofuel_hefa_others_share
            / 100
            * biofuel_hefa_others_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        self.df.loc[:, "biofuel_carbon_tax_hefa_others"] = biofuel_carbon_tax_hefa_others

        biofuel_mfsp_carbon_tax_supplement_hefa_others = (
            carbon_tax
            * biofuel_hefa_others_emission_factor
            / 1000000
            * lhv_biofuel
            * density_biofuel
        )
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_hefa_others"] = (
            biofuel_mfsp_carbon_tax_supplement_hefa_others
        )

        ### FT OTHERS
        biofuel_cost_ft_others = (
            biofuel_ft_others_mfsp
            / density_biofuel
            / lhv_biofuel
            * energy_consumption_biofuel
            * biofuel_ft_others_share
            / 100
            / 1e6
        )
        self.df.loc[:, "biofuel_cost_ft_others"] = biofuel_cost_ft_others

        biofuel_carbon_tax_ft_others = (
            energy_consumption_biofuel
            * biofuel_ft_others_share
            / 100
            * biofuel_ft_others_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        self.df.loc[:, "biofuel_carbon_tax_ft_others"] = biofuel_carbon_tax_ft_others

        biofuel_mfsp_carbon_tax_supplement_ft_others = (
            carbon_tax * biofuel_ft_others_emission_factor / 1000000 * lhv_biofuel * density_biofuel
        )
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_ft_others"] = (
            biofuel_mfsp_carbon_tax_supplement_ft_others
        )

        ### FT MSW
        biofuel_cost_ft_msw = (
            biofuel_ft_msw_mfsp
            / density_biofuel
            / lhv_biofuel
            * energy_consumption_biofuel
            * biofuel_ft_msw_share
            / 100
            / 1e6
        )
        self.df.loc[:, "biofuel_cost_ft_msw"] = biofuel_cost_ft_msw

        biofuel_carbon_tax_ft_msw = (
            energy_consumption_biofuel
            * biofuel_ft_msw_share
            / 100
            * biofuel_ft_msw_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        self.df.loc[:, "biofuel_carbon_tax_ft_msw"] = biofuel_carbon_tax_ft_msw

        biofuel_mfsp_carbon_tax_supplement_ft_msw = (
            carbon_tax * biofuel_ft_msw_emission_factor / 1000000 * lhv_biofuel * density_biofuel
        )
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_ft_msw"] = (
            biofuel_mfsp_carbon_tax_supplement_ft_msw
        )

        ### ATJ
        biofuel_cost_atj = (
            biofuel_atj_mfsp
            / density_biofuel
            / lhv_biofuel
            * energy_consumption_biofuel
            * biofuel_atj_share
            / 100
            / 1e6
        )
        self.df.loc[:, "biofuel_cost_atj"] = biofuel_cost_atj

        biofuel_carbon_tax_atj = (
            energy_consumption_biofuel
            * biofuel_atj_share
            / 100
            * biofuel_atj_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        self.df.loc[:, "biofuel_carbon_tax_atj"] = biofuel_carbon_tax_atj

        biofuel_mfsp_carbon_tax_supplement_atj = (
            carbon_tax * biofuel_atj_emission_factor / 1000000 * lhv_biofuel * density_biofuel
        )
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_atj"] = (
            biofuel_mfsp_carbon_tax_supplement_atj
        )

        # MEAN tax
        biofuel_mean_carbon_tax_per_l = (
            biofuel_mfsp_carbon_tax_supplement_hefa_fog.fillna(0)
            * biofuel_hefa_fog_share.fillna(0)
            / 100
            + biofuel_mfsp_carbon_tax_supplement_hefa_others.fillna(0)
            * biofuel_hefa_others_share.fillna(0)
            / 100
            + biofuel_mfsp_carbon_tax_supplement_ft_others.fillna(0)
            * biofuel_ft_others_share.fillna(0)
            / 100
            + biofuel_mfsp_carbon_tax_supplement_ft_msw.fillna(0)
            * biofuel_ft_msw_share.fillna(0)
            / 100
            + biofuel_mfsp_carbon_tax_supplement_atj.fillna(0) * biofuel_atj_share.fillna(0) / 100
        )

        self.df.loc[:, "biofuel_mean_carbon_tax_per_l"] = biofuel_mean_carbon_tax_per_l

        return (
            biofuel_cost_hefa_fog,
            biofuel_carbon_tax_hefa_fog,
            biofuel_mfsp_carbon_tax_supplement_hefa_fog,
            biofuel_cost_hefa_others,
            biofuel_carbon_tax_hefa_others,
            biofuel_mfsp_carbon_tax_supplement_hefa_others,
            biofuel_cost_ft_others,
            biofuel_carbon_tax_ft_others,
            biofuel_mfsp_carbon_tax_supplement_ft_others,
            biofuel_cost_ft_msw,
            biofuel_carbon_tax_ft_msw,
            biofuel_mfsp_carbon_tax_supplement_ft_msw,
            biofuel_cost_atj,
            biofuel_carbon_tax_atj,
            biofuel_mfsp_carbon_tax_supplement_atj,
            biofuel_mean_carbon_tax_per_l,
        )


class BiofuelMfspSimple(AeroMAPSModel):
    def __init__(self, name="biofuel_mfsp_simple", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        biofuel_hefa_fog_mfsp_simple_reference_years: list,
        biofuel_hefa_fog_mfsp_simple_reference_years_values: list,
        biofuel_hefa_others_mfsp_simple_reference_years: list,
        biofuel_hefa_others_mfsp_simple_reference_years_values: list,
        biofuel_ft_others_mfsp_simple_reference_years: list,
        biofuel_ft_others_mfsp_simple_reference_years_values: list,
        biofuel_ft_msw_mfsp_simple_reference_years: list,
        biofuel_ft_msw_mfsp_simple_reference_years_values: list,
        biofuel_atj_mfsp_simple_reference_years: list,
        biofuel_atj_mfsp_simple_reference_years_values: list,
        biofuel_hefa_fog_share: pd.Series,
        biofuel_hefa_others_share: pd.Series,
        biofuel_ft_others_share: pd.Series,
        biofuel_ft_msw_share: pd.Series,
        biofuel_atj_share: pd.Series,
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        """Biofuel MFSP (Minimal fuel selling price) estimates"""

        # HEFA FOG
        biofuel_hefa_fog_mfsp = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_fog_mfsp_simple_reference_years,
            biofuel_hefa_fog_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_fog_mfsp"] = biofuel_hefa_fog_mfsp

        # HEFA OTHERS
        biofuel_hefa_others_mfsp = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_others_mfsp_simple_reference_years,
            biofuel_hefa_others_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_others_mfsp"] = biofuel_hefa_others_mfsp

        # FT OTHERS
        biofuel_ft_others_mfsp = AeromapsInterpolationFunction(
            self,
            biofuel_ft_others_mfsp_simple_reference_years,
            biofuel_ft_others_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_others_mfsp"] = biofuel_ft_others_mfsp

        # FT MSW
        biofuel_ft_msw_mfsp = AeromapsInterpolationFunction(
            self,
            biofuel_ft_msw_mfsp_simple_reference_years,
            biofuel_ft_msw_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_msw_mfsp"] = biofuel_ft_msw_mfsp

        # ATJ
        biofuel_atj_mfsp = AeromapsInterpolationFunction(
            self,
            biofuel_atj_mfsp_simple_reference_years,
            biofuel_atj_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_atj_mfsp"] = biofuel_atj_mfsp

        # MEAN
        biofuel_mean_mfsp = (
            biofuel_hefa_fog_mfsp * biofuel_hefa_fog_share / 100
            + biofuel_hefa_others_mfsp * biofuel_hefa_others_share / 100
            + biofuel_ft_others_mfsp * biofuel_ft_others_share / 100
            + biofuel_ft_msw_mfsp * biofuel_ft_msw_share / 100
            + biofuel_atj_mfsp * biofuel_atj_share / 100
        )

        self.df.loc[:, "biofuel_mean_mfsp"] = biofuel_mean_mfsp

        # MARGINAL

        biofuel_marginal_mfsp = self.df.loc[
            :,
            [
                "biofuel_hefa_fog_mfsp",
                "biofuel_hefa_others_mfsp",
                "biofuel_ft_others_mfsp",
                "biofuel_ft_msw_mfsp",
                "biofuel_atj_mfsp",
            ],
        ].max(axis="columns")

        self.df.loc[:, "biofuel_marginal_mfsp"] = biofuel_marginal_mfsp

        return (
            biofuel_hefa_fog_mfsp,
            biofuel_hefa_others_mfsp,
            biofuel_ft_others_mfsp,
            biofuel_ft_msw_mfsp,
            biofuel_atj_mfsp,
            biofuel_mean_mfsp,
            biofuel_marginal_mfsp,
        )



import pandas as pd

from aeromaps.models.base import AeroMAPSModel, AeromapsInterpolationFunction


class PassengerAircraftNonOpCosts(AeroMAPSModel):
    def __init__(self, name="passenger_aircraft_noc", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        noc_reference_years: list,
        noc_reference_years_values: list,
    ) -> pd.Series:
        # Simple computation of airline non-operating costs (NOC)
        noc_prospective = AeromapsInterpolationFunction(
            self,
            noc_reference_years,
            noc_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "non_operating_cost_per_ask"] = noc_prospective
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "non_operating_cost_per_ask"] = self.df.loc[
                self.prospection_start_year, "non_operating_cost_per_ask"
            ]
        non_operating_cost_per_ask = self.df["non_operating_cost_per_ask"]
        return non_operating_cost_per_ask


class PassengerAircraftPassengerTax(AeroMAPSModel):
    def __init__(self, name="passenger_aircraft_passenger_tax", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        passenger_tax_reference_years: list,
        passenger_tax_reference_years_values: list,
    ) -> pd.Series:
        # Simple computation of airline non-operating costs (NOC)

        passenger_tax_prospective = AeromapsInterpolationFunction(
            self,
            passenger_tax_reference_years,
            passenger_tax_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "passenger_tax_per_ask"] = passenger_tax_prospective
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "passenger_tax_per_ask"] = self.df.loc[
                self.prospection_start_year, "passenger_tax_per_ask"
            ]
        passenger_tax_per_ask = self.df["passenger_tax_per_ask"]
        return passenger_tax_per_ask
